//+------------------------------------------------------------------+
//|                                                     MfsTrade.mqh |
//|                        Copyright 2017, Mukachi Financial Service |
//|                                               http://www.mfs.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2017, Mukachi Financial Service"
#property link      "http://www.mfs.com"
#property version   "1.00"
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#define MAX_RETRIES 5		// Max retries on error
#define RETRY_DELAY 2000	// Retry delay in ms

#include <MfsTrade/MfsTradeError.mqh>
//+------------------------------------------------------------------+
//| CTrade Class - Open, close_market and Modify Orders              |
//+------------------------------------------------------------------+
class CTrade
  {
private:
   MqlTradeRequest   m_request;

   bool              open_market(string symbol,ENUM_ORDER_TYPE type,
                                 double volume,double stop=0,
                                 double profit=0,string comment=NULL);

   bool              open_pending(string symbol,ENUM_ORDER_TYPE type,
                                  double volume,double price,double stop=0,
                                  double profit=0,double stop_limit=0,
                                  datetime expiration=0,string comment=NULL);

public:
   MqlTradeResult    result;

   bool              buy(string symbol,double volume,double stop=0,
                         double profit=0,string comment=NULL);

   bool              sell(string symbol,double volume,double stop=0,
                          double profit=0,string comment=NULL);

   bool              buy_stop(string symbol,double volume,double price,
                              double stop=0,double profit=0,
                              datetime expiration=0,
                              string comment=NULL);

   bool              sell_stop(string symbol,double volume,double price,
                               double stop=0,double profit=0,
                               datetime expiration=0,
                               string comment=NULL);

   bool              buy_limit(string symbol,double volume,double price,
                               double stop=0,double profit=0,
                               datetime expiration=0,
                               string comment=NULL);

   bool              sell_limit(string symbol,double volume,double price,
                                double stop=0,double profit=0,
                                datetime expiration=0,
                                string comment=NULL);

   bool              buy_stop_limit(string symbol,double volume,
                                    double price,double stop_limit,
                                    double stop=0,double profit=0,
                                    datetime expiration=0,
                                    string comment=NULL);

   bool              sell_stop_limit(string symbol,double volume,
                                     double price,double stop_limit,
                                     double stop=0,double profit=0,
                                     datetime expiration=0,
                                     string comment=NULL);

   bool              modify_market(string symbol,double stop,
                                   double profit=0);

   bool              modify_pending(ulong ticket,double price,
                                    double stop,double profit,
                                    datetime expiration=0);

   bool              delete_pending(ulong ticket);
   bool              close_market(string symbol,double volume=0,
                                  string comment=NULL);

   void              magic_number(ulong magic);
   void              deviation(ulong deviation);
   void              fill_type(ENUM_ORDER_TYPE_FILLING fill);
  };
//+------------------------------------------------------------------+
bool CTrade::open_market(string symbol,
                         ENUM_ORDER_TYPE type,
                         double volume,
                         double stop=0,
                         double profit=0,
                         string comment=NULL)
  {
   m_request.action=TRADE_ACTION_DEAL;
   m_request.symbol=symbol;
   m_request.type=type;
   m_request.sl=stop;
   m_request.tp=profit;
   m_request.comment=comment;

// calculate lot size
   double lots=0;
   long pos_type=WRONG_VALUE;

   if(PositionSelect(symbol))
     {
      lots=PositionGetDouble(POSITION_VOLUME);
      pos_type=PositionGetInteger(POSITION_TYPE);
     }

   if((type==ORDER_TYPE_BUY && pos_type==POSITION_TYPE_SELL) ||
      (type==ORDER_TYPE_SELL && pos_type==POSITION_TYPE_BUY))
     {
      m_request.volume=volume+lots;
     }
   else m_request.volume=volume;

// order loop
   int check_code = 0;
   int retry_count= 0;

   do
     {
      if(type==ORDER_TYPE_BUY)
         m_request.price=SymbolInfoDouble(symbol,SYMBOL_ASK);
      else if(type==ORDER_TYPE_SELL)
         m_request.price=SymbolInfoDouble(symbol,SYMBOL_BID);

      bool order_send=OrderSend(m_request,result);
      check_code=check_return_code(result.retcode);

      if(check_code==CHECK_RETCODE_OK) break;
      else if(check_code==CHECK_RETCODE_ERROR)
        {
         string err_desc=return_code_description(result.retcode);
         Alert("Open market order: Error ",
               result.retcode," - ",err_desc); break;
        }
      else // CHECK_RETCODE_RETRY
        {
         Print("Server error detected, retrying...");
         Sleep(RETRY_DELAY);
         retry_count++;
        }
     }
   while(retry_count<MAX_RETRIES);

   if(retry_count>=MAX_RETRIES)
     {
      string err_desc=return_code_description(result.retcode);
      Alert("Max retries exceeded: Error ",
            result.retcode," - ",err_desc);
     }

   string order_type=check_order_type(type);
   string err_desc=return_code_description(result.retcode);

   Print("Open ",order_type," order #",result.deal,": ",
         result.retcode," - ",err_desc,", Volume: ",
         result.volume,", Price: ",result.price,", Bid: ",
         result.bid,", Ask: ",result.ask);

   if(check_code==CHECK_RETCODE_OK)
     {
      Comment(order_type," position opened at ",
              result.price," on ",symbol);
      return(true);
     }
   else return(false);
  }
//+------------------------------------------------------------------+
bool CTrade::open_pending(string symbol,ENUM_ORDER_TYPE type,
                          double volume,double price,
                          double stop=0.000000,double profit=0.000000,
                          double stop_limit=0,datetime expiration=0,
                          string comment=NULL)
  {
   m_request.action=TRADE_ACTION_PENDING;
   m_request.symbol=symbol;
   m_request.type=type;
   m_request.sl=stop;
   m_request.tp=profit;
   m_request.comment=comment;
   m_request.price=price;
   m_request.volume=volume;
   m_request.stoplimit=stop_limit;

   if(expiration>0)
     {
      m_request.expiration= expiration;
      m_request.type_time = ORDER_TIME_SPECIFIED;
     }
   else m_request.type_time=ORDER_TIME_GTC;

// order loop
   int check_code = 0;
   int retry_count= 0;

   do
     {
      bool order_send=OrderSend(m_request,result);
      check_code=check_return_code(result.retcode);

      if(check_code==CHECK_RETCODE_OK) break;
      else if(check_code==CHECK_RETCODE_ERROR)
        {
         string err_desc=return_code_description(result.retcode);
         Alert("Open pending order: Error ",
               result.retcode," - ",err_desc); break;
        }
      else
        {
         Print("Server error detected, retrying...");
         Sleep(RETRY_DELAY);
         retry_count++;
        }
     }
   while(retry_count<MAX_RETRIES);

   if(retry_count>=MAX_RETRIES)
     {
      string err_desc=return_code_description(result.retcode);
      Alert("Max retries exceeded: Error ",
            result.retcode," - ",err_desc);
     }

   string order_type=check_order_type(type);
   string err_desc=return_code_description(result.retcode);

   Print("Open ",order_type," order #",result.order,": ",
         result.retcode," - ",err_desc,", Volume: ",
         result.volume,", Price: ",m_request.price,
         ", Bid: ",SymbolInfoDouble(symbol,SYMBOL_BID),
         ", Ask: ",SymbolInfoDouble(symbol,SYMBOL_ASK),
         ", SL: ",m_request.sl,", TP: ",m_request.tp,
         ", Stop Limit: ",m_request.stoplimit,
         ", Expiration: ",m_request.expiration);

   if(check_code==CHECK_RETCODE_OK)
     {
      Comment(order_type," order opened at ",
              m_request.price," on ",symbol);
      return(true);
     }
   else return(false);
  }
//+------------------------------------------------------------------+
bool CTrade::modify_market(string symbol,double stop,
                           double profit=0.000000)
  {
   m_request.action = TRADE_ACTION_SLTP;
   m_request.symbol = symbol;
   m_request.sl = stop;
   m_request.tp = profit;

// order loop
   int retry_count= 0;
   int check_code = 0;

   do
     {
      bool order_send=OrderSend(m_request,result);
      check_code=check_return_code(result.retcode);

      if(check_code==CHECK_RETCODE_OK) break;
      else if(check_code==CHECK_RETCODE_ERROR)
        {
         string err_desc=return_code_description(result.retcode);
         Alert("Modify position: Error ",
               result.retcode," - ",err_desc); break;
        }
      else
        {
         Print("Server error detected, retrying...");
         Sleep(RETRY_DELAY);
         retry_count++;
        }
     }
   while(retry_count<MAX_RETRIES);

   if(retry_count>=MAX_RETRIES)
     {
      string err_desc=return_code_description(result.retcode);
      Alert("Max retries exceeded: Error ",
            result.retcode," - ",err_desc);
     }

   string err_desc=return_code_description(result.retcode);

   Print("Modify position: ",result.retcode," - ",err_desc,
         ", SL: ",m_request.sl,", TP: ",m_request.tp,
         ", Bid: ",SymbolInfoDouble(symbol,SYMBOL_BID),
         ", Ask: ",SymbolInfoDouble(symbol,SYMBOL_ASK),
         ", Stop Level: ",SymbolInfoInteger(symbol,
         SYMBOL_TRADE_STOPS_LEVEL));

   if(check_code==CHECK_RETCODE_OK)
     {
      Comment("Position modified on ",symbol,", SL: ",
              m_request.sl,", TP: ",m_request.tp);
      return(true);
     }
   else return(false);
  }
//+------------------------------------------------------------------+
bool CTrade::modify_pending(ulong ticket,double price,
                            double stop,double profit,
                            datetime expiration=0)
  {
   m_request.action= TRADE_ACTION_MODIFY;
   m_request.order = ticket;
   m_request.sl = stop;
   m_request.tp = profit;

   bool order_select=OrderSelect(ticket);

   if(price>0) m_request.price=price;
   else m_request.price=OrderGetDouble(ORDER_PRICE_OPEN);

   if(expiration>0)
     {
      m_request.expiration= expiration;
      m_request.type_time = ORDER_TIME_SPECIFIED;
     }
   else m_request.type_time=ORDER_TIME_GTC;

// Order loop
   int retry_count= 0;
   int check_code = 0;

   do
     {
      bool order_send=OrderSend(m_request,result);
      check_code=check_return_code(result.retcode);

      if(check_code==CHECK_RETCODE_OK) break;
      else if(check_code==CHECK_RETCODE_ERROR)
        {
         string err_desc=return_code_description(result.retcode);
         Alert("Modify pending order: Error ",
               result.retcode," - ",err_desc); break;
        }
      else
        {
         Print("Server error detected, retrying...");
         Sleep(RETRY_DELAY);
         retry_count++;
        }
     }
   while(retry_count<MAX_RETRIES);

   if(retry_count>=MAX_RETRIES)
     {
      string err_desc=return_code_description(result.retcode);
      Alert("Max retries exceeded: Error ",
            result.retcode," - ",err_desc);
     }

   order_select=OrderSelect(ticket);
   string err_desc=return_code_description(result.retcode);

   Print("Modify pending order #",ticket,
         ": ",result.retcode," - ",err_desc,
         ", Price: ",OrderGetDouble(ORDER_PRICE_OPEN),
         ", SL: ",m_request.sl,", TP: ",m_request.tp,
         ", Expiration: ",m_request.expiration);

   if(check_code==CHECK_RETCODE_OK)
     {
      Comment("Pending order ",ticket," modified,",
              " Price: ",OrderGetDouble(ORDER_PRICE_OPEN),
              ", SL: ",m_request.sl,", TP: ",m_request.tp);
      return(true);
     }
   else return(false);
  }
//+------------------------------------------------------------------+
bool CTrade::close_market(string symbol,double volume=0.000000,
                          string comment=NULL)
  {
   m_request.action = TRADE_ACTION_DEAL;
   m_request.symbol = symbol;

   double close_lots=0;
   long open_type=WRONG_VALUE;
   long pos_id=0;

   if(PositionSelect(symbol)==true)
     {
      close_lots=PositionGetDouble(POSITION_VOLUME);
      open_type=PositionGetInteger(POSITION_TYPE);
      pos_id=PositionGetInteger(POSITION_IDENTIFIER);
      m_request.sl=PositionGetDouble(POSITION_SL);
      m_request.tp=PositionGetDouble(POSITION_TP);
     }
   else return(false);

   if(volume>close_lots || volume<=0)
      m_request.volume=close_lots;
   else m_request.volume=volume;

// Order loop
   int check_code = 0;
   int retry_count= 0;

   do
     {
      if(open_type==POSITION_TYPE_BUY)
        {
         m_request.type=ORDER_TYPE_SELL;
         m_request.price=SymbolInfoDouble(symbol,SYMBOL_BID);
        }
      else if(open_type==POSITION_TYPE_SELL)
        {
         m_request.type=ORDER_TYPE_BUY;
         m_request.price=SymbolInfoDouble(symbol,SYMBOL_ASK);
        }

      bool order_send=OrderSend(m_request,result);
      check_code=check_return_code(result.retcode);

      if(check_code==CHECK_RETCODE_OK) break;
      else if(check_code==CHECK_RETCODE_ERROR)
        {
         string err_desc=return_code_description(result.retcode);
         Alert("close_market position: Error ",
               result.retcode," - ",err_desc); break;
        }
      else
        {
         Print("Server error detected, retrying...");
         Sleep(RETRY_DELAY);
         retry_count++;
        }
     }
   while(retry_count<MAX_RETRIES);

   if(retry_count>=MAX_RETRIES)
     {
      string err_desc=return_code_description(result.retcode);
      Alert("Max retries exceeded: Error ",
            result.retcode," - ",err_desc);
     }

   string pos_type;
   if(open_type==POSITION_TYPE_BUY) pos_type="buy";
   else if(open_type==POSITION_TYPE_SELL) pos_type="sell";

   string err_desc=return_code_description(result.retcode);
   Print("close_market ",pos_type," position #",result.deal,
         ": ",result.retcode," - ",err_desc,", Volume: ",
         result.volume,", Price: ",result.price,", Bid: ",
         result.bid,", Ask: ",result.ask);

   if(check_code==CHECK_RETCODE_OK)
     {
      Comment(pos_type," position closed on ",
              symbol," at ",result.price);
      return(true);
     }
   else return(false);
  }
//+------------------------------------------------------------------+
bool CTrade::delete_pending(ulong ticket)
  {
   m_request.action= TRADE_ACTION_REMOVE;
   m_request.order = ticket;

// Order loop
   int retry_count= 0;
   int check_code = 0;

   do
     {
      bool order_send=OrderSend(m_request,result);
      check_code=check_return_code(result.retcode);

      if(check_code==CHECK_RETCODE_OK) break;
      else if(check_code==CHECK_RETCODE_ERROR)
        {
         string err_desc=return_code_description(result.retcode);
         Alert("delete_pending order: Error ",
               result.retcode," - ",err_desc); break;
        }
      else
        {
         Print("Server error detected, retrying...");
         Sleep(RETRY_DELAY);
         retry_count++;
        }
     }
   while(retry_count<MAX_RETRIES);

   if(retry_count>=MAX_RETRIES)
     {
      string err_desc=return_code_description(result.retcode);
      Alert("Max retries exceeded: Error ",
            result.retcode," - ",err_desc);
     }

   string err_desc=return_code_description(result.retcode);
   Print("delete_pending order #",ticket,
         ": ",result.retcode," - ",err_desc);

   if(check_code==CHECK_RETCODE_OK)
     {
      Comment("Pending order ",ticket," deleted");
      return(true);
     }
   else return(false);
  }
//+------------------------------------------------------------------+
bool CTrade::buy(string symbol,double volume,double stop=0.00000,
                 double profit=0.00000,string comment=NULL)
  {
   return(open_market(symbol,ORDER_TYPE_BUY,
          volume,stop,profit,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::sell(string symbol,double volume,double stop=0.00000,
                  double profit=0.00000,string comment=NULL)
  {
   return(open_market(symbol,ORDER_TYPE_SELL,
          volume,stop,profit,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::buy_limit(string symbol,double volume,double price,
                       double stop=0.00000,double profit=0.00000,
                       datetime expiration=0,string comment=NULL)
  {
   return(open_pending(symbol,ORDER_TYPE_BUY_LIMIT,
          volume,price,stop,profit,0,expiration,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::sell_limit(string symbol,double volume,double price,
                        double stop=0.00000,double profit=0.00000,
                        datetime expiration=0,string comment=NULL)
  {
   return(open_pending(symbol,ORDER_TYPE_SELL_LIMIT,
          volume,price,stop,profit,0,expiration,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::buy_stop(string symbol,double volume,double price,
                      double stop=0.000000,double profit=0.000000,
                      datetime expiration=0,string comment=NULL)
  {
   return(open_pending(symbol,ORDER_TYPE_BUY_STOP,
          volume,price,stop,profit,0,expiration,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::sell_stop(string symbol,double volume,double price,
                       double stop=0.00000,double profit=0.00000,
                       datetime expiration=0,string comment=NULL)
  {
   return(open_pending(symbol,ORDER_TYPE_SELL_STOP,
          volume,price,stop,profit,0,expiration,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::buy_stop_limit(string symbol,double volume,
                            double price,double stop_limit,
                            double stop=0.00000,double profit=0.00000,
                            datetime expiration=0,string comment=NULL)
  {
   return(open_pending(symbol,ORDER_TYPE_BUY_STOP_LIMIT,
          volume,price,stop,profit,stop_limit,expiration,comment));
  }
//+------------------------------------------------------------------+
bool CTrade::sell_stop_limit(string symbol,double volume,
                             double price,double stop_limit,
                             double stop=0.00000,double profit=0.00000,
                             datetime expiration=0,string comment=NULL)
  {
   return(open_pending(symbol,ORDER_TYPE_SELL_STOP_LIMIT,
          volume,price,stop,profit,stop_limit,expiration,comment));
  }
//+------------------------------------------------------------------+
void CTrade::magic_number(ulong magic)
  {
   m_request.magic=magic;
  }
//+------------------------------------------------------------------+
void CTrade::deviation(ulong deviation)
  {
   m_request.deviation=deviation;
  }
//+------------------------------------------------------------------+
void CTrade::fill_type(ENUM_ORDER_TYPE_FILLING fill)
  {
   m_request.type_filling=fill;
  }
//+------------------------------------------------------------------+
int check_return_code(uint return_code)
  {
   int status;
//---
   switch(return_code)
     {
      case TRADE_RETCODE_REQUOTE:
      case TRADE_RETCODE_CONNECTION:
      case TRADE_RETCODE_PRICE_CHANGED:
      case TRADE_RETCODE_TIMEOUT:
      case TRADE_RETCODE_PRICE_OFF:
      case TRADE_RETCODE_REJECT:
      case TRADE_RETCODE_ERROR:

         status=CHECK_RETCODE_RETRY;

         break;

      case TRADE_RETCODE_DONE:
      case TRADE_RETCODE_DONE_PARTIAL:
      case TRADE_RETCODE_PLACED:
      case TRADE_RETCODE_NO_CHANGES:

         status=CHECK_RETCODE_OK;

         break;

      default: status=CHECK_RETCODE_ERROR;
     }
//---
   return(status);
  }
//+------------------------------------------------------------------+
//| Stop Loss & Take Profit Calculation Functions                    |
//+------------------------------------------------------------------+
double buy_stop_loss(string symbol,
                     int stop_points,
                     double price=0)
  {
   if(stop_points <= 0) return(0);

   double order_price;
   if(price>0) order_price=price;
   else order_price=SymbolInfoDouble(symbol,SYMBOL_ASK);

   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double stop_loss=order_price-(stop_points*point);

   long digits=SymbolInfoInteger(symbol,SYMBOL_DIGITS);
   stop_loss=NormalizeDouble(stop_loss,(int)digits);

   return(stop_loss);
  }
//+------------------------------------------------------------------+
double sell_stop_loss(string symbol,
                      int stop_points,
                      double price=0)
  {
   if(stop_points <= 0) return(0);

   double order_price;
   if(price>0) order_price=price;
   else order_price=SymbolInfoDouble(symbol,SYMBOL_BID);

   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double stop_loss=order_price+(stop_points*point);

   long digits=SymbolInfoInteger(symbol,SYMBOL_DIGITS);
   stop_loss=NormalizeDouble(stop_loss,(int)digits);

   return(stop_loss);
  }
//+------------------------------------------------------------------+
double buy_take_profit(string symbol,int profit_points,double price=0)
  {
   if(profit_points <= 0) return(0);

   double order_price;
   if(price>0) order_price=price;
   else order_price=SymbolInfoDouble(symbol,SYMBOL_ASK);

   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double take_profit=order_price+(profit_points*point);

   long digits= SymbolInfoInteger(symbol,SYMBOL_DIGITS);
   take_profit= NormalizeDouble(take_profit,(int)digits);

   return(take_profit);
  }
//+------------------------------------------------------------------+
double sell_take_profit(string symbol,int profit_points,double price=0)
  {
   if(profit_points <= 0) return(0);

   double order_price;
   if(price>0) order_price=price;
   else order_price=SymbolInfoDouble(symbol,SYMBOL_BID);

   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double take_profit=order_price -(profit_points*point);

   long digits= SymbolInfoInteger(symbol,SYMBOL_DIGITS);
   take_profit= NormalizeDouble(take_profit,(int)digits);

   return(take_profit);
  }
//+------------------------------------------------------------------+
//| Stop Level Verification                                          |
//+------------------------------------------------------------------+
bool check_above_stop_level(string symbol,double price,int points=10)
  {
   double current_price=SymbolInfoDouble(symbol,SYMBOL_ASK);
   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double stop_level=SymbolInfoInteger(symbol,
                                       SYMBOL_TRADE_STOPS_LEVEL)*point;
   double stop_price=current_price+stop_level;
   double buffer_points=points*point;

   if(price>=(stop_price+buffer_points)) return(true);
   else return(false);
  }
//+------------------------------------------------------------------+
bool check_below_stop_level(string symbol,double price,int points=10)
  {
   double current_price=SymbolInfoDouble(symbol,SYMBOL_BID);
   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double stop_level=SymbolInfoInteger(symbol,SYMBOL_TRADE_STOPS_LEVEL)*point;
   double stop_price=current_price-stop_level;
   double buffer_points=points*point;

   if(price <= stop_price - buffer_points) return(true);
   else return(false);
  }
//+------------------------------------------------------------------+
double adjust_above_stop_level(string symbol,double price,int points=10)
  {
   double current_price=SymbolInfoDouble(symbol,SYMBOL_ASK);
   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double stop_level=SymbolInfoInteger(symbol,SYMBOL_TRADE_STOPS_LEVEL)*point;
   double stop_price=current_price+stop_level;
   double buffer_points=points*point;

   if(price > stop_price + buffer_points) return(price);
   else
     {
      double new_price=stop_price+buffer_points;
      Print("Price adjusted above stop level to "+
            DoubleToString(new_price));
      return(new_price);
     }
  }
//+------------------------------------------------------------------+
double adjust_below_stop_level(string symbol,double price,int points=10)
  {
   double current_price=SymbolInfoDouble(symbol,SYMBOL_BID);
   double point=SymbolInfoDouble(symbol,SYMBOL_POINT);
   double stop_level=SymbolInfoInteger(symbol,SYMBOL_TRADE_STOPS_LEVEL)*point;
   double stop_price=current_price-stop_level;
   double buffer_points=points*point;

   if(price < stop_price - buffer_points) return(price);
   else
     {
      double new_price=stop_price-buffer_points;
      Print("Price adjusted below stop level to "+
            DoubleToString(new_price));
      return(new_price);
     }
  }
//+------------------------------------------------------------------+
//| Position Information                                             |
//+------------------------------------------------------------------+
string position_comment(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetString(POSITION_COMMENT));
   else return(NULL);
  }
//+------------------------------------------------------------------+
long position_type(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetInteger(POSITION_TYPE));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
long position_identifier(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetInteger(POSITION_IDENTIFIER));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
double position_open_price(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetDouble(POSITION_PRICE_OPEN));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
long position_open_time(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetInteger(POSITION_TIME));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
double position_volume(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetDouble(POSITION_VOLUME));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
double position_stop_loss(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetDouble(POSITION_SL));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
double position_take_profit(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetDouble(POSITION_TP));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
double position_profit(string symbol=NULL)
  {
   if(symbol==NULL) symbol=_Symbol;
   if(PositionSelect(symbol))
      return(PositionGetDouble(POSITION_PROFIT));
   else return(WRONG_VALUE);
  }
//+------------------------------------------------------------------+
//| Miscellaneous Functions & Enumerations                           |
//+------------------------------------------------------------------+
enum ENUM_CHECK_RETCODE
  {
   CHECK_RETCODE_OK,
   CHECK_RETCODE_ERROR,
   CHECK_RETCODE_RETRY
  };
//+------------------------------------------------------------------+
string check_order_type(ENUM_ORDER_TYPE type)
  {
   string order_type;
//---
   if(type==ORDER_TYPE_BUY)
      order_type="buy";
   else if(type==ORDER_TYPE_SELL)
      order_type="sell";
   else if(type==ORDER_TYPE_BUY_STOP)
      order_type="buy stop";
   else if(type==ORDER_TYPE_BUY_LIMIT)
      order_type="buy limit";
   else if(type==ORDER_TYPE_SELL_STOP)
      order_type="sell stop";
   else if(type==ORDER_TYPE_SELL_LIMIT)
      order_type="sell limit";
   else if(type==ORDER_TYPE_BUY_STOP_LIMIT)
      order_type="buy stop limit";
   else if(type==ORDER_TYPE_SELL_STOP_LIMIT)
      order_type="sell stop limit";
   else order_type="invalid order type";
//---
   return(order_type);
  }
//+------------------------------------------------------------------+
